#undef MOVL
#undef MOVQ
#undef MOVAPS
#undef SFENCE
#ifdef NT
#define MOVL	"	movnti	" /* needs sse2 */
#define MOVQ	"	movntq	" /* needs sse or mmxext */
#define MOVAPS	"	movntps	" /* needs sse */
#define SFENCE	"	sfence\n" /* needs sse or mmxext */
#else
#define MOVL	"	movl	"
#define MOVQ	"	movq	" /* needs mmx */
#define MOVAPS	"	movaps	" /* needs sse */
#define SFENCE
#endif

//NB:
// PREFETCH[W] needs 3dnow
// PREFETCH{T0,T1,T2,NTA} needs sse

#undef ALIGN16
#ifdef ALIGN
#define ALIGN16	"	.align	16\n"
#else
#define ALIGN16
#endif


#define SSE_START(cr0) do { \
	preempt_disable(); \
	cr0 = read_cr0(); \
	clts(); \
	} while(0)

#define SSE_END(cr0) do { \
	write_cr0(cr0); \
	preempt_enable(); \
	} while(0)

void FUNC(zero_page_sse)(void *page)
{
	unsigned char xmm_save[16];
	unsigned int cr0;

	SSE_START(cr0);
	asm volatile(
		"	movups	%%xmm0, %0\n"
		"	xorps	%%xmm0, %%xmm0\n"
		: : "m" (*xmm_save)
	);
	asm volatile(
			ALIGN16
		"1:\n"
			MOVAPS "%%xmm0, (%0)\n"
			MOVAPS "%%xmm0, 16(%0)\n"
		"	cmpl	%0, %1\n"
			MOVAPS "%%xmm0, 32(%0)\n"
			MOVAPS "%%xmm0, 48(%0)\n"
		"	leal	64(%0), %0\n"
		"	ja	1b\n"
		: : "r" (page), "r" ((char*)page+(PAGE_SIZE-64))
		: "memory"
	);
	asm volatile(
		"	movups	%0, %%xmm0\n"
			SFENCE
		: : "m" (*xmm_save) : "memory"
	);
	SSE_END(cr0);
}

void FUNC(copy_page_sse)(void *to, const void *from)
{
	unsigned char xmm_save[64];
	unsigned int cr0;

	SSE_START(cr0);
	asm volatile(
		"	movups	%%xmm0, (%0)\n"
		"	movups	%%xmm1, 16(%0)\n"
		"	movups	%%xmm2, 32(%0)\n"
		"	movups	%%xmm3, 48(%0)\n"
		: : "r" (xmm_save)
	);
#ifdef PREFETCH
	asm volatile (
		"1:	prefetcht0 (%0)\n"
		"	prefetcht0 64(%0)\n"
		"	prefetcht0 2*64(%0)\n"
		"	prefetcht0 3*64(%0)\n"
		"	prefetcht0 4*64(%0)\n"
		"2:\n"
		".section .fixup, \"ax\"\n"
		"3:	movw	$0x1AEB, 1b\n" /* jmp on 26 bytes */
		"	jmp	2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.align	4\n"
		"	.long	1b, 3b\n"
		".previous"
		: : "r" (from)
	);
#endif
	asm volatile(
			ALIGN16
		"1:\n"
#ifdef PREFETCH
		"	prefetcht0 5*64(%0)\n"
#endif
		"2:	movaps	(%0), %%xmm0\n"
		"	movaps	16(%0), %%xmm1\n"
		"	movaps	32(%0), %%xmm2\n"
		"	movaps	48(%0), %%xmm3\n"
		/* ? "	leal	64(%0), %0\n" */
		"	addl	$64, %0\n"
			MOVAPS "%%xmm0, (%1)\n"
			MOVAPS "%%xmm1, 16(%1)\n"
		"	cmpl	%0, %2\n"
			MOVAPS "%%xmm2, 32(%1)\n"
			MOVAPS "%%xmm3, 48(%1)\n"
		"	leal	64(%1), %1\n"
		"	ja	1b\n"
#ifdef PREFETCH
		"	cmpl	%0, %3\n"
		"	ja	2b\n"
		".section .fixup, \"ax\"\n"
		"3:	movw	$0x05EB, 1b\n"	/* jmp 7 bytes ahead */
		"	jmp	2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.align	4\n"
		"	.long	1b, 3b\n"
		".previous"
#endif
		: : "r" (from), "r" (to)
#ifdef PREFETCH
		    , "r" ((char*)from+(PAGE_SIZE-5*64))
#endif
		    , "r" ((char*)from+PAGE_SIZE)
		: "memory"
	);
	asm volatile(
		"	movups	(%0), %%xmm0\n"
		"	movups	16(%0), %%xmm1\n"
		"	movups	32(%0), %%xmm2\n"
		"	movups	48(%0), %%xmm3\n"
			SFENCE
		: : "r" (xmm_save) : "memory"
	);
	SSE_END(cr0);
}

void FUNC(zero_page_mmx)(void *page)
{
	kernel_fpu_begin();
	__asm__ __volatile__ (
		"	pxor	%%mm0, %%mm0\n"
			ALIGN16
		"1:\n"
			MOVQ	"%%mm0, (%0)\n"
			MOVQ	"%%mm0, 8(%0)\n"
			MOVQ	"%%mm0, 16(%0)\n"
			MOVQ	"%%mm0, 24(%0)\n"
		"	cmpl	%0, %1\n"
			MOVQ	"%%mm0, 32(%0)\n"
			MOVQ	"%%mm0, 40(%0)\n"
			MOVQ	"%%mm0, 48(%0)\n"
			MOVQ	"%%mm0, 56(%0)\n"
		"	leal	64(%0), %0\n"
		"	ja	1b\n"
			SFENCE
		: : "r" (page), "r" ((char*)page+(PAGE_SIZE-64))
		: "memory"
	);
	kernel_fpu_end();
}

void FUNC(copy_page_mmx)(void *to, const void *from)
{
#ifdef PREFETCH
	asm volatile (
		"1:	prefetch (%0)\n"
		"	prefetch 64(%0)\n"
		"	prefetch 2*64(%0)\n"
		"	prefetch 3*64(%0)\n"
		"	prefetch 4*64(%0)\n"
		"2:\n"
		".section .fixup, \"ax\"\n"
		"3:	movw	$0x1AEB, 1b\n" /* jmp on 26 bytes */
		"	jmp	2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.align	4\n"
		"	.long	1b, 3b\n"
		".previous"
		: : "r" (from)
	);
#endif
	kernel_fpu_begin();
	asm volatile(
			ALIGN16
		"1:\n"
#ifdef PREFETCH
		"	prefetch 5*64(%0)\n"
#endif
		"2:	movq	(%0), %%mm0\n"
		"	movq	8(%0), %%mm1\n"
		"	movq	16(%0), %%mm2\n"
		"	movq	24(%0), %%mm3\n"
		"	movq	32(%0), %%mm4\n"
		"	movq	40(%0), %%mm5\n"
		"	movq	48(%0), %%mm6\n"
		"	movq	56(%0), %%mm7\n"
		/* ? "	leal	64(%0), %0\n" */
		"	addl	$64, %0\n"
			MOVQ	"%%mm0, (%1)\n"
			MOVQ	"%%mm1, 8(%1)\n"
			MOVQ	"%%mm2, 16(%1)\n"
			MOVQ	"%%mm3, 24(%1)\n"
		"	cmpl	%0, %2\n"
			MOVQ	"%%mm4, 32(%1)\n"
			MOVQ	"%%mm5, 40(%1)\n"
			MOVQ	"%%mm6, 48(%1)\n"
			MOVQ	"%%mm7, 56(%1)\n"
		"	leal	64(%1), %1\n"
		"	ja	1b\n"
#ifdef PREFETCH
		"	cmpl	%0, %3\n"
		"	ja	2b\n"
		".section .fixup, \"ax\"\n"
		"3:	movw	$0x05EB, 1b\n"	/* jmp 7 bytes ahead */
		"	jmp	2b\n"
		".previous\n"
		".section __ex_table,\"a\"\n"
		"	.align	4\n"
		"	.long	1b, 3b\n"
		".previous"
#endif
			SFENCE
		: : "r" (from), "r" (to)
#ifdef PREFETCH
		    , "r" ((char*)from+(PAGE_SIZE-5*64))
#endif
		    , "r" ((char*)from+PAGE_SIZE)
		: "memory"
	);
	kernel_fpu_end();
}

void FUNC(zero_page_gp)(void *page)
{
	int d;
	asm volatile(
		"	xorl	%0, %0\n"
			ALIGN16
		"1:\n"
			MOVL	"%0, (%1)\n"
			MOVL	"%0, 4(%1)\n"
			MOVL	"%0, 8(%1)\n"
			MOVL	"%0, 12(%1)\n"
		"	cmpl	%1, %2\n"
			MOVL	"%0, 16(%1)\n"
			MOVL	"%0, 20(%1)\n"
			MOVL	"%0, 24(%1)\n"
			MOVL	"%0, 28(%1)\n"
		"	leal	32(%1), %1\n"
		"	ja	1b\n"
			SFENCE
		: "=&r" (d)
		: "r" (page), "r" ((char*)page+(PAGE_SIZE-32))
		: "memory"
	);
}

void FUNC(copy_page_gp)(void *to, const void *from)
{
	int d2,d3,d4;
	asm volatile(
			ALIGN16
		"1:\n"
#ifdef PREFETCH
		"	movl	32(%3), %2\n"
#endif
		"2:	movl	(%3), %0\n"
		"	movl	4(%3), %1\n"
		"	movl	8(%3), %2\n"
			MOVL	"%0, (%4)\n"
			MOVL	"%1, 4(%4)\n"
			MOVL	"%2, 8(%4)\n"
		"	movl	12(%3), %0\n"
		"	movl	16(%3), %1\n"
		"	movl	20(%3), %2\n"
			MOVL	"%0, 12(%4)\n"
			MOVL	"%1, 16(%4)\n"
			MOVL	"%2, 20(%4)\n"
		"	movl	24(%3), %0\n"
		"	movl	28(%3), %1\n"
		/* ? "	leal	32(%3), %3\n" */
		"	addl	$32, %3\n"
			MOVL	"%0, 24(%4)\n"
		"	cmpl	%3, %5\n"
			MOVL	"%1, 28(%4)\n"
		"	leal	32(%4), %4\n"
		"	ja	1b\n"
#ifdef PREFETCH
		"	je	2b\n"
#endif
		: "=&r" (d2), "=&r" (d3), "=&r" (d4)
#ifdef PREFETCH
		: "r" (from), "r" (to), "r" ((char*)from+(PAGE_SIZE-32))
#else
		: "r" (from), "r" (to), "r" ((char*)from+PAGE_SIZE)
#endif
		: "memory"
	);
}

#undef FUNC
